/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SWO_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void SurfaceWrapOp::initialize(void)
{
	//* NOTE inherits attributes from SurfaceBindOp

	catalogRemove("PivotPoint");

	catalog<String>("BindPrefix","hint")=
			"Prefix of attribute names on input surface describing"
			" where each point on the input surface is attached to"
			" the associated driver.";

	catalog<Real>("Distance Bias")=1.0;
	catalog<Real>("Distance Bias","min")=0.0;
	catalog<Real>("Distance Bias","high")=1.0;
	catalog<Real>("Distance Bias","max")=10.0;
	catalog<String>("Distance Bias","hint")=
			"Dominance of closer faces towards weighting";

	catalog<bool>("Debug Bind")=false;
	catalog<bool>("Debug Bind","joined")=true;
	catalog<String>("Debug Bind","hint")=
			"Draws lines over the output showing the attachments."
			"  This feedback may only be visible is this node is"
			" \"selected\" and/or \"displayed\".";

	catalog<I32>("Debug Index")=-1;
	catalog<I32>("Debug Index","min")=-1;
	catalog<I32>("Debug Index","high")=1000;
	catalog<I32>("Debug Index","max")=1000000;
	catalog<String>("Debug Index","enabler")="Debug Bind";
	catalog<String>("Debug Index","hint")=
			"Index of element to examine (-1 to see them all).";

	catalog< sp<Component> >("Deformed Surface");
	catalog<String>("Deformed Surface","label")="Driver Deformed Surface";

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","recycle")=true;

	m_spDebugGroup=new DrawMode();
	m_spDebugGroup->setGroup("debug");
}

void SurfaceWrapOp::handle(Record& a_rSignal)
{
#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle\n");
#endif

	//* output curves is copy of input surface
	if(!accessOutput(m_spOutputAccessible,a_rSignal)) return;

	if(catalog<BWORD>("Debug Bind"))
	{
		if(!accessGuide(m_spDrawDebug,a_rSignal)) return;
	}
	else
	{
		m_spDrawDebug=NULL;
	}

	if(!access(m_spInputAccessible,"Input Surface")) return;

	//* original driver
	const String driverGroup=catalog<String>("DriverGroup");
	if(!access(m_spDriver,"Driver Surface",driverGroup)) return;

	m_spDriver->setRefinement(0);
	m_spDriver->prepareForSearch();

	//* deformed driver
	if(!access(m_spDeformed,"Deformed Surface",driverGroup)) return;

	m_spDeformed->setRefinement(0);
	m_spDeformed->prepareForSearch();

	const bool bindPrimitives=catalog<bool>("Bind Primitives");

	m_element=bindPrimitives? e_primitive: e_point;
	m_attribute=bindPrimitives? e_vertices: e_position;

	sp<SurfaceAccessorI> spOutputElements;
	if(!access(spOutputElements,m_spOutputAccessible,
			m_element,m_attribute)) return;

	const U32 elementCount=spOutputElements->count();

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->pushDrawMode(m_spDebugGroup);
	}

	U32 runCount=elementCount;
	m_fragmented=catalog<bool>("Bind Fragments");
	m_spFragmentAccessor=NULL;
	if(m_fragmented)
	{
		if(access(m_spFragmentAccessor,m_spOutputAccessible,
				m_element,m_attribute))
		{
			const String fragmentAttribute=
					catalog<String>("FragmentAttr");
			m_spFragmentAccessor->fragmentWith(
					SurfaceAccessibleI::Element(m_element),fragmentAttribute);
			runCount=m_spFragmentAccessor->fragmentCount();
			if(runCount<1)	//* was (runCount<2), but there can be one fragment
			{
				catalog<String>("warning")=
						"fragmentation failed to divide surface -> reverting";
				m_fragmented=FALSE;
				runCount=elementCount;
			}
		}
		else
		{
			catalog<String>("warning")=
					"fragmentation access failed -> ignoring";
			m_fragmented=FALSE;
			runCount=elementCount;
		}
	}

	setFullRange(0,runCount-1);
	OperatorThreaded::handle(a_rSignal);

	if(m_spDrawDebug.isValid())
	{
		m_spDrawDebug->popDrawMode();
	}

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle done\n");
#endif
}

void SurfaceWrapOp::run(I32 a_id,Range a_range)
{
	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;

	const Real distanceBias=catalog<Real>("Distance Bias");
	const I32 debugIndex=catalog<I32>("Debug Index");

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run running thread %d [%d,%d]\n",a_id,start,end);
#endif

	sp<SurfaceAccessorI> spInputElements;
	if(!access(spInputElements,m_spInputAccessible,
			m_element,m_attribute)) return;

	sp<SurfaceAccessorI> spOutputElements;
	if(!access(spOutputElements,m_spOutputAccessible,
			m_element,m_attribute)) return;

	std::vector< sp<SurfaceAccessorI> > outputFaces;
	std::vector< sp<SurfaceAccessorI> > outputBarys;
	const I32 bindings=accessBindings(m_spOutputAccessible,
			outputFaces,outputBarys,FALSE);
	if(bindings<1)
	{
		return;
	}

	const bool spiderWrap=catalog<bool>("Spider");
	const BWORD singular=
			(m_fragmented && m_element==SurfaceAccessibleI::e_point);

	sp<SurfaceAccessorI::FilterI> spFilter;

	for(U32 index=start;index<=end;index++)
	{
		I32 filterCount=1;
		U32 pointCount=0;
		if(m_fragmented)
		{
			const String fragment=m_spFragmentAccessor->fragment(index);
			if(!m_spFragmentAccessor->filterWith(fragment,spFilter) ||
					spFilter.isNull())
			{
				continue;
			}
			filterCount=spFilter->filterCount();
			if(filterCount<1)
			{
				continue;
			}
			if(singular)
			{
				pointCount+=filterCount;
			}
			else
			{
				for(I32 filterIndex=0;filterIndex<filterCount;filterIndex++)
				{
					const U32 filtered=spFilter->filter(filterIndex);
					pointCount+=spInputElements->subCount(filtered);
				}
			}
		}
		else
		{
			pointCount=spInputElements->subCount(index);
		}

		if(pointCount<1)
		{
			continue;
		}

		SpatialVector point[pointCount];
		SpatialVector result[pointCount];
		Real totalWeight[pointCount];

		U32 pointIndex=0;
		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const U32 filtered=
					m_fragmented? spFilter->filter(filterIndex): index;
			const U32 subCount=singular? 1: spInputElements->subCount(filtered);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				point[pointIndex]=
						spInputElements->spatialVector(filtered,subIndex);
				set(result[pointIndex]);
				totalWeight[pointIndex]=0.0;
				pointIndex++;
			}
		}

		const U32 keyIndex=m_fragmented? spFilter->filter(0): index;

		for(U32 binding=0;binding<bindings;binding++)
		{
			const I32 face=outputFaces[binding]->integer(keyIndex);
			if(face<0)
			{
				continue;
			}

			SpatialBary barycenter(0.33,0.33);
			if(!binding && !spiderWrap)
			{
				barycenter=outputBarys[binding]->spatialVector(keyIndex);
			}
			const SpatialTransform xformDriver=
					m_spDriver->sample(face,barycenter);
			const SpatialTransform xformDeformed=
					m_spDeformed->sample(face,barycenter);
			SpatialTransform invDriver;
			invert(invDriver,xformDriver);
			const SpatialTransform conversion=invDriver*xformDeformed;

			const Real distance=
					magnitude(xformDriver.translation()-point[0]);
			const Real weight=1.0/(distance*distanceBias+1e-9);

			for(pointIndex=0;pointIndex<pointCount;pointIndex++)
			{
				SpatialVector transformed;
				transformVector(conversion,point[pointIndex],transformed);
				result[pointIndex]+=weight*transformed;
				totalWeight[pointIndex]+=weight;
			}

			if(m_spDrawDebug.isValid() && (debugIndex<0 || debugIndex==index))
			{
				SpatialVector transformed;
				transformVector(conversion,point[0],transformed);

				const Color yellow(1.0,1.0,0.0,1.0);
				SpatialVector line[2];

				line[0]=transformed;
				line[1]=xformDeformed.translation();

				SAFEGUARD;
				m_spDrawDebug->drawLines(line,NULL,2,DrawI::e_strip,
						false,&yellow);
			}
		}

		pointIndex=0;
		for(U32 filterIndex=0;filterIndex<filterCount;filterIndex++)
		{
			const U32 filtered=
					m_fragmented? spFilter->filter(filterIndex): index;
			const U32 subCount=singular? 1: spInputElements->subCount(filtered);
			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				if(totalWeight[pointIndex]>0.0)
				{
					spOutputElements->set(filtered,subIndex,
							result[pointIndex]/totalWeight[pointIndex]);
				}

				pointIndex++;
			}
		}
	}

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run done running thread %d\n",a_id);
#endif
}
